%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Etvs Lornd University.
%%% Portions created  by Etvs  Lornd University are  Copyright 2009,
%%% Etvs Lornd University. All Rights Reserved.

%%% @author Judit Koszegi <kojqaai@inf.elte.hu>

-module(web_helper).
-vsn("$Rev: 8858 $ ").

-export([start_yaws/1]).
-export([get_index_yaws_folder/0]).
-export([convert_list_to_integer/1, string_to_ip/1]).

-include("ui.hrl").


%%% ============================================================================
%%% Start, configure and stop yaws web server

%% @doc Configure and start yaws web server.
start_yaws(["from_script", YPath, YName, YPort, YListen]) ->
    try
        YawsPathProp = {yaws_path,YPath},
        YawsNameProp = {yaws_name,YName},
        YawsPortProp = {yaws_port, convert_list_to_integer(YPort)},
        YawsListenProp = {yaws_listen, string_to_ip(YListen)},
        YawsIndexFunModuleProp = {yaws_index_fun_module, ?MODULE},
        YawsIndexFunNameProp = {yaws_index_fun_name, get_index_yaws_folder},
        AllProp = [YawsNameProp,YawsPortProp,YawsListenProp,
                   YawsIndexFunModuleProp,YawsIndexFunNameProp],
        case YPath of
            "no_path" -> start_yaws(AllProp);
            _ -> start_yaws([YawsPathProp|AllProp])
        end
    catch
        RefErr ->
            io:format(?Error:error_text(RefErr) ++ "~n"),
            exit("Cannot start yaws.")
    end;
start_yaws(Proplist) ->
    case (proplists:get_value(yaws_path,Proplist)) of
        undefined -> ok;
        YawsDir ->
            case code:add_path(YawsDir) of
                true -> ok;
                {error,_} ->
                    throw(?RefError(file_notdir,[YawsDir]))
            end
    end,
    case code:ensure_loaded(yaws) of
        {error,_} -> throw(?RefErr0r(yaws_not_loaded));
            _ -> ok
    end,
    Port = proplists:get_value(yaws_port,Proplist,8001),
    validate_port_number(Port),
    Listen = proplists:get_value(yaws_listen,Proplist,{0,0,0,0}),
    validate_ip_tuple(Listen),
    Name0 = proplists:get_value(yaws_name,Proplist,"refactorErl"),
    Name = case is_atom(Name0) of
               true -> atom_to_list(Name0);
               false -> Name0
           end,
    validate_name(Name),
    YawsIndexFunModule = proplists:get_value(yaws_index_fun_module,
                                             Proplist,web_helper),
    YawsIndexFunName = proplists:get_value(yaws_index_fun_name,Proplist,
                                           get_index_yaws_folder),
    YawsAppMod = proplists:get_value(yaws_app_mod,Proplist,no_appmod),
    Nitrogen = proplists:get_value(nitrogen_prop,Proplist,no_nitrogen),
    YawsPostSize = proplists:get_value(yaws_partial_post_size,Proplist,1024),
    
    GconfList = case Nitrogen of
                    no_nitrogen -> [{id, Name},
                                    {servername, Name}];
                    with_nitrogen ->[
                                     {ebin_dir,
                [?NITRO_CORE:get_nitrogen_site_dir()++"/ebin"]},
                                     {include_dir,
                [?NITRO_CORE:get_nitrogen_site_dir()++"/include"]},
                                     {id, Name},
                                     {servername,Name}]
                end,

    SconfList = 
    case YawsAppMod of
        no_appmod ->[{docroot, apply(YawsIndexFunModule,YawsIndexFunName,[])},
                     {port, Port},
                     {listen, Listen}];
        _ ->        [{docroot, apply(YawsIndexFunModule,YawsIndexFunName,[])},
                     {port, Port},
                     {listen, Listen},
                     {appmods, YawsAppMod},
                    {partial_post_size,YawsPostSize}]
    end,
    yaws:start_embedded(apply(YawsIndexFunModule,YawsIndexFunName,[]), 
        SconfList,GconfList).


validate_port_number(Port) ->
    case is_number(Port) of
        true -> ok;
        false ->throw(?RefError(port_format_error,[io_lib:format("~p",[Port])]))
    end.

validate_ip_tuple(Listen) ->
    Str = io_lib:format("~p",[Listen]),
    case Listen of
        {A,B,C,D} ->
            case is_number(A) andalso is_number(B) andalso
                is_number(C) andalso is_number(D) of
                true -> ok;
                false -> throw(?RefError(ip_format_error,[Str]))
            end;
        _ -> throw(?RefError(ip_format_error,[Str]))
    end.

validate_name(Name) ->
    case is_string(Name) of
        true -> ok;
        false ->throw(?RefError(name_format_error,[io_lib:format("~p",[Name])]))
    end.

is_char(Ch) ->
    if Ch < 0 -> false;
       Ch > 255 -> false;
       true -> true
    end.

is_string(Str) ->
    case is_list(Str) of
        false -> false;
        true -> lists:all(fun is_char/1, Str)
    end.

get_index_yaws_folder() ->
    Path0 = filename:split(filename:dirname(code:which(web_helper))),
    Path =
        filename:join([
                       filename:join(
                         lists:sublist(Path0,length(Path0)-1)),
                      "web"]),
    Path.

convert_list_to_integer(String) ->
    try
        list_to_integer(String)
    catch _A:_B ->
            throw(?RefError(list_to_integer_error,[String]))
    end.

take_and_drop(L) ->
    {convert_list_to_integer(lists:takewhile(fun(X) -> not(X == $.) end,L)),
     (lists:dropwhile(fun(X) -> not(X == $.) end,L))}.


string_to_ip(YPort) ->
    L1 = YPort,
    {Number1,L2} = take_and_drop(L1),
    case L2 of
        [] ->  throw(?RefError(ip_format_error,[YPort]));
        _ ->
            {Number2,L3} = take_and_drop(tl(L2)),
            case L3 of
                [] -> throw(?RefError(ip_format_error,[YPort]));
                _ ->  {Number3,L4} = take_and_drop(tl(L3)),
                      case L4 of
                          [] -> throw(?RefError(ip_format_error,[YPort]));
                          _ ->  {Number4,_L5} = take_and_drop(tl(L4)),
                                {Number1,Number2,Number3,Number4}
                      end
            end
    end.
